static char nftversion[] = {
	1,1,6,
	0
};
static char nftbuildstamp[] = {
	((uint64_t)MAKE_STAMP >> 56) & 0xff,
	((uint64_t)MAKE_STAMP >> 48) & 0xff,
	((uint64_t)MAKE_STAMP >> 40) & 0xff,
	((uint64_t)MAKE_STAMP >> 32) & 0xff,
	((uint64_t)MAKE_STAMP >> 24) & 0xff,
	((uint64_t)MAKE_STAMP >> 16) & 0xff,
	((uint64_t)MAKE_STAMP >> 8) & 0xff,
	((uint64_t)MAKE_STAMP >> 0) & 0xff,
};
